from stockpyle._base import BaseDictionaryStore


class ShoveStore(BaseDictionaryStore):
    """Represents a store that places all objects in a Shove (see http://pypi.python.org/pypi/shove)"""
        
    def __init__(self, shove=None, shoveuri=None, polymorphic=False, lifetime_cb=None):
        if shove is not None and shoveuri is not None:
            raise ValueError("you can only provide either a Shove object or a Shove URI, not both")
        elif shove is not None:
            self.__shove = shove
        elif shoveuri is not None:
            from shove import Shove
            self.__shove = Shove(shoveuri)
        else:
            raise ValueError("you must provide either a Shove object or a Shove URI to create a ShoveStore")
        super(ShoveStore, self).__init__(dictionary=self.__shove, polymorphic=polymorphic, lifetime_cb=lifetime_cb)



from unittest import TestCase
from stockpyle._helpers import KeyValueHelper


class DummyObject(object):
    
    __stockpyle_keys__ = [("foo", "bar")]
    foo = 1
    bar = 2


class DummyObjectSubclass(DummyObject):
    
    __stockpyle_keys__ = ["zap"]
    zap = "xyz"


class DummyObjectSubSubclassWithoutStockpyleKeys(DummyObjectSubclass):
    pass


class DummyObjectSubSubSubclass(DummyObjectSubSubclassWithoutStockpyleKeys):
    
    __stockpyle_keys__ = [("foo", "blarg")]
    blarg = "abc"


class KeyValueHelperTestCase(TestCase):
    
    def setUp(self):
        self.__module_name_for_dummy_objects = str(DummyObject.__module__)
    
    def test_generates_proper_verbose_keys(self):
        
        kvh = KeyValueHelper()
        key = kvh.generate_verbose_lookup_key(DummyObject, {"xyz": "foo", "abc": 123})
        self.assertEqual(key, "stockpyle:%s.DummyObject(abc=123;xyz=foo)" % self.__module_name_for_dummy_objects)
    
    def test_generates_proper_terse_keys(self):
        
        kvh = KeyValueHelper()
        key = kvh.generate_terse_lookup_key(DummyObject, {"xyz": "foo", "abc": 123})
        self.assertEqual(key, "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, hash("abc=123;xyz=foo")))
    
    def test_generates_proper_verbose_keys_list_when_nonpolymorphic(self):
        
        kvh = KeyValueHelper()
        
        # test dummy object
        keys = kvh.generate_all_verbose_lookup_keys(DummyObject())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, "bar=2;foo=1")
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_verbose_lookup_keys(DummyObjectSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, "zap=xyz")
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_verbose_lookup_keys(DummyObjectSubSubclassWithoutStockpyleKeys())
        expected_set = set([
            "stockpyle:%s.DummyObjectSubSubclassWithoutStockpyleKeys(%s)" % (self.__module_name_for_dummy_objects, "zap=xyz")
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_verbose_lookup_keys(DummyObjectSubSubSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObjectSubSubSubclass(%s)" % (self.__module_name_for_dummy_objects, "blarg=abc;foo=1")
            ])
        self.assertEqual(set(keys), expected_set)
    
    def test_generates_proper_terse_keys_list_when_nonpolymorphic(self):
        
        kvh = KeyValueHelper()
        
        # test dummy object
        keys = kvh.generate_all_terse_lookup_keys(DummyObject())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, hash("bar=2;foo=1"))
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_terse_lookup_keys(DummyObjectSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, hash("zap=xyz"))
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_terse_lookup_keys(DummyObjectSubSubclassWithoutStockpyleKeys())
        expected_set = set([
            "stockpyle:%s.DummyObjectSubSubclassWithoutStockpyleKeys(%s)" % (self.__module_name_for_dummy_objects, hash("zap=xyz"))
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_terse_lookup_keys(DummyObjectSubSubSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObjectSubSubSubclass(%s)" % (self.__module_name_for_dummy_objects, hash("blarg=abc;foo=1"))
            ])
        self.assertEqual(set(keys), expected_set)
    
    def test_generates_proper_verbose_keys_list_when_polymorphic(self):
        
        kvh = KeyValueHelper(polymorphic=True)
        
        # test dummy object
        keys = kvh.generate_all_verbose_lookup_keys(DummyObject())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, "bar=2;foo=1")
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_verbose_lookup_keys(DummyObjectSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, "bar=2;foo=1"),
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, "zap=xyz"),
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_verbose_lookup_keys(DummyObjectSubSubclassWithoutStockpyleKeys())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, "bar=2;foo=1"),
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, "zap=xyz"),
            "stockpyle:%s.DummyObjectSubSubclassWithoutStockpyleKeys(%s)" % (self.__module_name_for_dummy_objects, "zap=xyz"),
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_verbose_lookup_keys(DummyObjectSubSubSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, "bar=2;foo=1"),
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, "zap=xyz"),
            "stockpyle:%s.DummyObjectSubSubclassWithoutStockpyleKeys(%s)" % (self.__module_name_for_dummy_objects, "zap=xyz"),
            "stockpyle:%s.DummyObjectSubSubSubclass(%s)" % (self.__module_name_for_dummy_objects, "blarg=abc;foo=1"),
            ])
        self.assertEqual(set(keys), expected_set)
    
    def test_generates_proper_terse_keys_list_when_polymorphic(self):
        
        kvh = KeyValueHelper(polymorphic=True)
        
        # test dummy object
        keys = kvh.generate_all_terse_lookup_keys(DummyObject())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, hash("bar=2;foo=1")),
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_terse_lookup_keys(DummyObjectSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, hash("bar=2;foo=1")),
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, hash("zap=xyz")),
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_terse_lookup_keys(DummyObjectSubSubclassWithoutStockpyleKeys())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, hash("bar=2;foo=1")),
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, hash("zap=xyz")),
            "stockpyle:%s.DummyObjectSubSubclassWithoutStockpyleKeys(%s)" % (self.__module_name_for_dummy_objects, hash("zap=xyz")),
            ])
        self.assertEqual(set(keys), expected_set)
        
        # test subclass
        keys = kvh.generate_all_terse_lookup_keys(DummyObjectSubSubSubclass())
        expected_set = set([
            "stockpyle:%s.DummyObject(%s)" % (self.__module_name_for_dummy_objects, hash("bar=2;foo=1")),
            "stockpyle:%s.DummyObjectSubclass(%s)" % (self.__module_name_for_dummy_objects, hash("zap=xyz")),
            "stockpyle:%s.DummyObjectSubSubclassWithoutStockpyleKeys(%s)" % (self.__module_name_for_dummy_objects, hash("zap=xyz")),
            "stockpyle:%s.DummyObjectSubSubSubclass(%s)" % (self.__module_name_for_dummy_objects, hash("blarg=abc;foo=1")),
            ])
        self.assertEqual(set(keys), expected_set)
    
        


from paver.defaults import Bunch

options(
    setup=Bunch(
        name="stockpyle",
        packages=["stockpyle", "stockpyle.ext"],
        version="0.0.9",
        license="BSD",
        author="Matt Pizzimenti",
        author_email="mjpizz+stockpyle@gmail.com",
        url="http://pypi.python.org/pypi/stockpyle/",
        install_requires = [],
        classifiers = [
            "Development Status :: 3 - Alpha",
            "Programming Language :: Python",
            "License :: OSI Approved :: BSD License",
            "Intended Audience :: Developers",
            "Operating System :: OS Independent",
            "Natural Language :: English",
            "Topic :: Software Development",
            "Topic :: System :: Distributed Computing",
            "Topic :: Software Development :: Object Brokering",
            "Topic :: Database :: Front-Ends",
        ],
        description="stockpyle allows the creation of write-through storage for object caching and persistence",
        long_description='''\
Overview
===========

The central concept of stockpyle is the *Store* objects.  You can find a variety of these objects
in stockpyle.stores:

* ProcessMemoryStore
* ChainedStore
* MemcacheStore - http://pypi.python.org/pypi/python-memcached
* ShoveStore - http://pypi.python.org/pypi/shove
* SqlAlchemyStore - http://pypi.python.org/pypi/SQLAlchemy
    
Stores can perform 6 operations on objects:

* put
* get
* delete
* batch_put
* batch_get
* batch_delete

Every stockpyle Store can handle any Python object that exposes the property *__stockpyle_keys__*.
This property lists the names of each attribute or attribute-tuple that `uniquely` identifies
that object.  Read the quickstart below for more details.

Quickstart
==========

Basics
------

The fastest way to get up and running is to create an in-memory Store::
    
    from stockpyle.stores import ProcessMemoryStore
    
    store = ProcessMemoryStore()

Now lets create a Python object that can be stored in stockpyle.  This object
has an attribute 'key' that identifies it uniquely::
    
    class Foo(object):
        
        __stockpyle_keys__ = ["key"]
        
        def __init__(self, key):
            self.key = key

We can now store a Foo object in our Store::
    
    f = Foo('my_unique_key_123')
    store.put(f)

And we can retrieve it from the Store::
    
    f = store.get(Foo, {'key': 'my_unique_key_123'})

We can also delete it from the Store::

    store.delete(f)

Manipulating batches of objects is also trivial::
    
    f1 = Foo('my_key_1')
    f2 = Foo('my_key_2')
    f3 = Foo('my_key_3')
    
    store.batch_put([f1, f2, f3])
    
    o1, o2, o3 = store.batch_get(Foo, [
        {'key': 'my_key_1'},
        {'key': 'my_key_2'},
        {'key': 'my_key_3'},
        ])
    
    assert((f1, f2, f3) == (o1, o2, o3))
    
    store.batch_delete([f1, f2, f3])

Note: if you do a *batch_get* and one of the keys is null, it still gets returned as None::
    
    o1, o2, o3 = store.batch_get(Foo, [
        {'key': 'my_key_1'},
        {'key': 'nonexistent_key'},
        {'key': 'my_key_3'},
        ])
    
    assert(o2 is None)

And that's basics of using stockpyle.  No matter what Store object
you use, you can use this API to store, retrieve, and delete objects.

Chained Storage (e.g. write-through caching)
--------------------------------------------

Persistent storage is often expensive to access, making caching a necessity.
However, maintaining cache consistency is difficult and error-prone.  The stockpyle
API provides a consistent way to perform this caching, using *ChainedStore*.

ChainedStore takes an ordered list of stores and treats them as a write-through
cache.  This example creates a write-through cache for SQLAlchemy-backed database objects::

    fast_cache = ProcessMemoryStore()
    dist_cache = MemcacheStore(servers=["localhost:11211"])
    persistent_store = SqlAlchemyStore(uri="sqlite:///data.sqlite")
    
    chained_store = ChainedStore([fast_cache, dist_cache, persistent_store])

In this example, the fast cache (in-memory) will always be attempted first for
retrievals.  The distributed cache (memcached) will be attempted second.  If these
stores fail to find the object, the database store (SQLAlchemy) will be queried for
the persistent object.

During puts, gets, and deletes, stockpyle will keep the upper layers (the caches)
populated with the latest data from lower layers (the persistent database).

Bugs
====

This is still an early-stage project.  Please file bugs on Google Code:

http://code.google.com/p/stockpyle/

Advanced
========

Alternate Keys
--------------

The __stockpyle_keys__ attribute can contain more than just one unique attribute.
It can also contain tuples of attribute names::
    
    class Foo(object):
        
        __stockpyle_keys__ = ["key", ("zap", "bar")]
        
        def __init__(self, key, zap, bar):
            self.key = key
            self.zap = zap
            self.bar = bar

Now, Foo is indexed by both 'key', *and* the combined attributes 'zap' and 'bar'::
    
    f = Foo(key=1, zap=2, bar=3)
    
    store.put(f)
    
    o1 = store.get(Foo, {'key': 1})
    o2 = store.get(Foo, {'zap': 2, 'bar': 3})
    
    assert(f == o1 == o2)

Cache Expiration
----------------

Some stockpyle stores support the concept of object expiration:

* ProcessMemoryStore
* MemcachedStore
* ShoveStore

In all three stores, you can specify a "lifetime-callback", which takes
a Python object as an argument and returns a lifetime value for that object.

Lifetime can be expressed in 3 ways:

* integer number of seconds
* Python timedelta
* Python datetime (absolute expiration time)

This example allows Foo objects to live for 20 seconds, and Bar objects to
live for 5 minutes::

    def my_lifetime_cb(obj):
        
        if isinstance(obj, Foo):
            return 20
        
        elif isinstance(obj, Bar):
            return 5*60
    
    store = ProcessMemoryStore(lifetime_cb=my_lifetime_cb)

Note that you can specify different lifetime callbacks for each your stores.
For example, you may want to expire in-memory caches quickly to for more frequent
re-sync with the lower layers in a ChainedStore.

Polymorphic Storage
-------------------

Some stockpyle stores support the concept of storing objects in a way
that can be queried polymorphically:

* ProcessMemoryStore
* MemcachedStore
* ShoveStore

By default, objects are not stored polymorphically in these stores because
it duplicates the locations that a single object is stored.  However, you may
want to be able to query subclasses out of a cache::

    class Automobile(object):
        
        __stockpyle_keys__ = ["license_plate"]
        
        def __init__(self, license_plate):
            self.license_plate = license_plate
    
    class Car(Automobile):
        pass
    
    
    class Truck(Automobile):
        pass

If you create a store that supports polymorphic queries::
    
    store = ProcessMemoryStore(polymorphic=True)

Then you can store a Car::
    
    car = Car(license_plate="1234 XYZ")
    store.put()

And retrieve it as either a Car or an automobile::
    
    car = store.get(Car, {'license_plate': '1234 XYZ'})
    automobile = store.get(Automobile, {'license_plate': '1234 XYZ'})
    
    assert(car == automobile)

Note that there is a tradeoff in storage duplication when doing this.  Every Car
object is indexed as a Car and as an Automobile (2 locations in storage).

Memcache Integration
--------------------

In order to enable MemcacheStore, you will need to install the python-memcached module:

http://pypi.python.org/pypi/python-memcached

The MemcacheStore supports polymorphic storage and object expiration.  The native memcache API is used to support
object expiration and batched operations.

Shove Integration
-----------------

In order to enable ShoveStore, you will need to install the shove module:

http://pypi.python.org/pypi/shove

The Shove API provides a very broad array of storage backends with a dictionary-style interface,
which work well in stockpyle. This example uses the shove backend for Amazon S3::
    
    from stockpyle.stores import ShoveStore
    
    s3_store = ShoveStore(shoveuri="s3://s3key:s3secret@your_s3_bucket_name")

ShoveStores support polymorphic storage and object expiration.  Every object stored in the Shove is stored at
a human-readable key, as a tuple of (obj, expiredate).

SQLAlchemy Integration
----------------------

In order to enable SqlAlchemyStore, you will need to install the SQLAlchemy module:

http://pypi.python.org/pypi/SQLAlchemy

Typically, in SQLAlchemy you will use a session to store and retrieve your mapped objects::

    # create the object
    obj = MyMappedObject(foobar=123)
    
    # store it
    session.save(obj)
    session.commit()
    
    # retrieve it
    obj_retrieved = session.query(MyMappedObject).filter(MyMappedObject.foobar == 123).first()

You can do the same thing through a SqlAlchemyStore::
    
    # create the store
    store = SqlAlchemyStore(uri="sqlite:///data.sqlite")
    
    # create the object
    obj = MyMappedObject(foobar=123)
    
    # store it
    store.put(obj)
    
    # retrieve it
    obj_retrieved = store.get(MyMappedObject, {'foobar': 123})

Now you can put this SqlAlchemyStore in a ChainedStore for write-through caching::
    
    # use MemcacheStore as a cache layer
    mc = MemcacheStore(servers=["localhost:11211"])
    sa = SqlAlchemyStore(uri="sqlite:///data.sqlite")
    store = ChainedStore([mc, sa])
    
    # store it, which causes the object to be written-through
    # the MemcacheStore
    store.put(obj)
    
    # retrieve it, except this time it comes from
    # the MemcacheStore instead of the database
    obj_retrieved = store.get(MyMappedObject, {'id': obj.id})

*What if you want to access the SQLAlchemy session directly?*  This is a common
case, since obviously stockpyle does not provide the rich querying API that
SQLAlchemy has.  To ensure that your stockpyle cache does not get out of sync
with queries done in your own SQLAlchemy sessions, simply add a
StockpyleSessionExtension to your session object::
    
    from sqlalchemy.orm import sessionmaker
    from stockpyle.stores import MemcacheStore
    from stockpyle.ext.sqlalchemy import StockpyleSessionExtension

    mc = MemcacheStore(servers=["localhost:11211"])
    ext = StockpyleSessionExtension(store=mc)
    my_session = sessionmaker(extension=ext)

Now, whenever *my_session* performs an insert, update, or delete on an object,
it will also perform the corresponding puts and deletes in the MemcacheStore.

Objects that you get from SqlAlchemyStore are bound to the internal session
object on that store.  Keep this in mind when you are using your own session,
since you will have to *merge()* the stockpyle-retrieved object into your session::
    
    # get an object from SqlAlchemyStore
    obj = sa_store.get(MyMappedObject, {'id': 'foobar123'})
    
    # change some information
    obj.name = "bleh"
    
    # if you want to use this object in your SQLAlchemy
    # session, you must merge the object first since the
    # object is currently bound to the internal session
    # inside the SqlAlchemyStore
    my_session.merge(obj)
    my_session.commit()

Defining a custom Store
-----------------------
Although stockpile contains most of the stores you will need to support
any storage backend (especially through Shove integration), you may
want to implement your own Store.  The stockpyle library provides two base
classes to define your own Stores:

* BaseStore (for implementing your own store from scratch)
* BaseDictionaryStore (for implementing your own store that has a dictionary-style interface)

To make your own store, simply create a subclass of either of these stores and implement
the 6 storage methods (put, get, delete, batch_put, batch_get, batch_delete).

Please browse the source code for detailed examples of how to implement your own store.
''',
        zip_safe = False,
        )
    )

@task
@needs(['generate_setup', 'minilib', 'setuptools.command.sdist'])
def sdist():
    """Overrides sdist to make sure that our setup.py is generated."""
    pass



from time import sleep
from datetime import datetime, timedelta
from sqlalchemy import engine_from_config, Column
from sqlalchemy.types import Integer, String
from sqlalchemy.ext.declarative import declarative_base


_DUMMY_EXPIRE_TIME = 1


DeclarativeBase = declarative_base()


class Foobar(DeclarativeBase):
    
    __tablename__ = "foobars"
    
    zap = Column(Integer, primary_key=True)
    foo = Column(Integer)
    bar = Column(Integer)
    
    __stockpyle_keys__ = [("foo", "bar"), ("bar", "zap"), "zap"]
    
    def __init__(self, foo, bar, zap):
        self.foo = foo
        self.bar = bar
        self.zap = zap
    
    def __eq__(self, other):
        return other and (other.foo, other.bar == self.foo, self.bar)


class StoreFixture(object):
    
    def _raise_not_implemented(self):
        raise NotImplementedError("all StoreFixtures must implement the 'store' property")
    
    store = property(_raise_not_implemented)
    
    def test_can_store_objects(self):
        obj = Foobar(1,2,3)
        try:
            self.store.put(obj)
        except Exception, e:
            self.fail("should be able to put objects (%s)" % str(e))
    
    def test_can_store_and_retrieve_by_key_foo_and_bar(self):
        obj = Foobar(1,2,3)
        self.store.put(obj)
        obj_r = self.store.get(Foobar, {"foo": 1, "bar": 2})
        self.assertEqual(obj_r, obj)
    
    def test_can_store_and_retrieve_by_key_bar_and_zap(self):
        obj = Foobar(1,2,3)
        self.store.put(obj)
        obj_r = self.store.get(Foobar, {"bar": 2, "zap": 3})
        self.assertEqual(obj_r, obj)
    
    def test_can_store_and_retrieve_by_key_zap(self):
        obj = Foobar(1,2,3)
        self.store.put(obj)
        obj_r = self.store.get(Foobar, {"zap": 3})
        self.assertEqual(obj_r, obj)
    
    def test_can_store_and_delete(self):
        obj = Foobar(1,2,3)
        self.store.put(obj)

        obj_r = self.store.get(Foobar, {"zap": 3})
        self.assertEqual(obj_r, obj)

        self.store.delete(obj)
        obj_r = self.store.get(Foobar, {"zap": 3})
        self.assertEqual(obj_r, None)
        obj_r = self.store.get(Foobar, {"foo": 1, "bar": 2})
        self.assertEqual(obj_r, None)
        obj_r = self.store.get(Foobar, {"bar": 2, "zap": 3})
        self.assertEqual(obj_r, None)
    
    def test_can_store_multiple_objects_and_retrieve_in_batch(self):
        
        obj1 = Foobar(1,2,3)
        obj2 = Foobar(4,5,6)
        
        self.store.put(obj1)
        self.store.put(obj2)
        
        obj1_r, obj2_r = self.store.batch_get(Foobar, [
            {"foo": 1, "bar": 2}, #obj1
            {"foo": 4, "bar": 5}, #obj2
            ])
        
        self.assertEqual(obj1_r, obj1)
        self.assertEqual(obj2_r, obj2)
        
        obj1_r, obj2_r = self.store.batch_get(Foobar, [
            {"bar": 2, "zap": 3}, #obj1
            {"bar": 5, "zap": 6}, #obj2
            ])
        
        self.assertEqual(obj1_r, obj1)
        self.assertEqual(obj2_r, obj2)
        
    def test_can_do_batch_put(self):
        
        obj1 = Foobar(1,2,3)
        obj2 = Foobar(4,5,6)
        
        self.store.batch_put([obj1, obj2])
        
        obj1_r, obj2_r = self.store.batch_get(Foobar, [
            {"foo": 1, "bar": 2}, #obj1
            {"foo": 4, "bar": 5}, #obj2
            ])
        
        self.assertEqual(obj1_r, obj1)
        self.assertEqual(obj2_r, obj2)
        
    def test_can_do_batch_delete_on_original_objects(self):
        
        obj1 = Foobar(1,2,3)
        obj2 = Foobar(4,5,6)
        
        self.store.put(obj1)
        self.store.put(obj2)
        
        obj1_r, obj2_r = self.store.batch_get(Foobar, [
            {"foo": 1, "bar": 2}, #obj1
            {"foo": 4, "bar": 5}, #obj2
            ])
        
        self.assertEqual(obj1_r, obj1)
        self.assertEqual(obj2_r, obj2)
        
        # now delete them
        self.store.batch_delete([obj1, obj2])
        
        obj1_r, obj2_r = self.store.batch_get(Foobar, [
            {"foo": 1, "bar": 2}, #obj1
            {"foo": 4, "bar": 5}, #obj2
            ])
        
        self.assertEqual(obj1_r, None)
        self.assertEqual(obj2_r, None)
        
    def test_can_do_batch_delete_on_retrieved_objects(self):
        
        obj1 = Foobar(1,2,3)
        obj2 = Foobar(4,5,6)
        
        self.store.put(obj1)
        self.store.put(obj2)
        
        obj1_r, obj2_r = self.store.batch_get(Foobar, [
            {"foo": 1, "bar": 2}, #obj1
            {"foo": 4, "bar": 5}, #obj2
            ])
        
        self.assertEqual(obj1_r, obj1)
        self.assertEqual(obj2_r, obj2)
        
        # now delete them
        self.store.batch_delete([obj1_r, obj2_r])
        
        obj1_r, obj2_r = self.store.batch_get(Foobar, [
            {"foo": 1, "bar": 2}, #obj1
            {"foo": 4, "bar": 5}, #obj2
            ])
        
        self.assertEqual(obj1_r, None)
        self.assertEqual(obj2_r, None)
    

class ExpiringStoreFixture(object):
    
    def lifetime_cb(self, obj):
        if hasattr(obj.__class__, "__stockpyle_unittest_lifetime__"):
            return obj.__class__.__stockpyle_unittest_lifetime__
        else:
            return None

    def test_can_store_and_expire_later(self):
        
        Foobar.__stockpyle_unittest_lifetime__ = _DUMMY_EXPIRE_TIME
        try:

            obj = Foobar(1,2,3)
            self.store.put(obj)
        
            obj_r = self.store.get(Foobar, {"zap": 3})
            self.assertEqual(obj, obj_r)
        
            # wait for expiration and check again
            sleep(_DUMMY_EXPIRE_TIME)
        
            obj_r = self.store.get(Foobar, {"zap": 3})
            self.assertEqual(obj_r, None)

        finally:
            Foobar.__stockpyle_unittest_lifetime__ = None
    
    def test_can_store_and_expire_later_with_datetime(self):
        
        d = datetime.today() + timedelta(seconds=_DUMMY_EXPIRE_TIME)
        Foobar.__stockpyle_unittest_lifetime__ = d
        
        try:

            obj = Foobar(1,2,3)
            self.store.put(obj)
        
            obj_r = self.store.get(Foobar, {"zap": 3})
            self.assertEqual(obj, obj_r)
        
            # wait for expiration and check again
            sleep(_DUMMY_EXPIRE_TIME)
        
            obj_r = self.store.get(Foobar, {"zap": 3})
            self.assertEqual(obj_r, None)

        finally:
            Foobar.__stockpyle_unittest_lifetime__ = None
    
    def test_can_store_and_expire_later_with_timedelta(self):
        
        td = timedelta(seconds=_DUMMY_EXPIRE_TIME)
        Foobar.__stockpyle_unittest_lifetime__ = td
        
        try:
            obj = Foobar(1,2,3)
            self.store.put(obj)
        
            obj_r = self.store.get(Foobar, {"zap": 3})
            self.assertEqual(obj, obj_r)
        
            # wait for expiration and check again
            sleep(_DUMMY_EXPIRE_TIME)
        
            obj_r = self.store.get(Foobar, {"zap": 3})
            self.assertEqual(obj_r, None)

        finally:
            Foobar.__stockpyle_unittest_lifetime__ = None



import os
import sys
from unittest import TestCase
import stockpyle._sqlalchemy


class SqlAlchemyMissingTestCase(TestCase):
    
    def setUp(self):
        # make sure that SQLAlchemy is not in the path
        self.__original_path = sys.path
        sys.path = [os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))]
    
    def tearDown(self):
        # restore the original sys path
        import sys
        sys.path = self.__original_path
    
    def test_session_extension_throws_exception(self):
        from stockpyle._sqlalchemy import StockpyleSessionExtension
        self.assertRaises(StandardError, lambda: StockpyleSessionExtension())



from unittest import TestCase
from tests.fixture_store import StoreFixture, ExpiringStoreFixture
from stockpyle._procmem import ProcessMemoryStore


class ProcessMemoryStoreTestCase(StoreFixture, ExpiringStoreFixture, TestCase):
    
    store = property(lambda self: self.__store)
    
    def setUp(self):
        self.__store = ProcessMemoryStore(lifetime_cb=self.lifetime_cb)
    


from unittest import TestCase
from stockpyle._base import BaseStore


class ProcessMemoryStoreTestCase(TestCase):
    
    def test_must_implement_storage_methods(self):
        
        
        class MyStore(BaseStore):
            pass
        
        
        class Foo(object):
            pass
            
        
        ms = MyStore()
        dummyobj = Foo()
        
        self.assertRaises(NotImplementedError, lambda: ms.put(dummyobj))
        self.assertRaises(NotImplementedError, lambda: ms.batch_put([dummyobj]))
        self.assertRaises(NotImplementedError, lambda: ms.delete(dummyobj))
        self.assertRaises(NotImplementedError, lambda: ms.batch_delete([dummyobj]))
        self.assertRaises(NotImplementedError, lambda: ms.get(Foo, {}))
        self.assertRaises(NotImplementedError, lambda: ms.batch_get(Foo, [{}, {}]))
    


from stockpyle._base import BaseStore, BaseDictionaryStore
from stockpyle._procmem import ProcessMemoryStore
from stockpyle._memcache import MemcacheStore
from stockpyle._shove import ShoveStore
from stockpyle._sqlalchemy import SqlAlchemyStore
from stockpyle._chained import ChainedStore



from stockpyle._base import BaseStore


class AppEngineStore(object):
    """represents a storage container on the Google App Engine DataStore"""
        
    def put(self, obj):
        from google.appengine.ext import db
        if isinstance(obj, db.Model):
            db.put(obj)
        
    def batch_put(self, objs):
        from google.appengine.ext import db
        ds_objs = [o for o in objs if isinstance(o, db.Model)]
        db.put(ds_objs)
    
    def delete(self, obj):
        from google.appengine.ext import db
        if isinstance(obj, db.Model):
            db.delete(obj)
    
    def batch_delete(self, objs):
        from google.appengine.ext import db
        ds_objs = [o for o in objs if isinstance(o, db.Model)]
        db.delete(ds_objs)
    
    def get(self, klass, key):
        from google.appengine.ext import db
        if issubclass(klass, db.Model):
            query = db.Query(klass)
            for k,v in key.iteritems():
                query.filter("%s=" % k, v)
            return query.get()
    
    def batch_get(self, klass, keys):
        # TODO: is there a more efficient way?
        return [self.get(klass, k) for k in keys]
            



from stockpyle._base import BaseDictionaryStore


class ProcessMemoryStore(BaseDictionaryStore):
    """Represents an in-process memory store"""
        
    def __init__(self, polymorphic=False, lifetime_cb=None):
        self.__dictionary = {}
        super(ProcessMemoryStore, self).__init__(dictionary=self.__dictionary, polymorphic=polymorphic, lifetime_cb=lifetime_cb)



from stockpyle._base import BaseStore


class ChainedStore(BaseStore):
    """Represents a chain of any arbitrary BaseStores.  Data will be written *through*
    all of these stores in order.  Data will be retrieved by attempting the first
    store, then the 2nd store, then the 3rd store, etc. until one of the stores returns
    a hit.  Hits are then stored forwards in all of the previous stores that had
    previously missed, so that next time an 'earlier' store can respond faster.
    
    Ideally, the stores would be ordered fastest-to-slowest.  A common use case for
    a ChainedStore is to have a persistent store with a multi-level cache front:
    
    fast_cache = ProcessMemoryStore()
    dist_cache = MemcacheStore(servers=["localhost:11711"])
    persistent_store = SqlAlchemyStore(uri="sqlite:///:memory:")
    
    chained_store = ChainedStore([fast_cache, dist_cache, persistent_store])
    """
    
    def __init__(self, stores=None):
        super(ChainedStore, self).__init__()
        if not stores:
            raise ValueError("you didn't specify any internal stores for this ChainedStore to operate on")
        self.__stores = stores
    
    def put(self, obj):
        
        # write-through
        for store in self.__stores:
            store.put(obj)
    
    def batch_put(self, objs):
        
        # write-through
        for store in self.__stores:
            store.batch_put(objs)
    
    def delete(self, obj):
        
        # delete-through
        for store in self.__stores:
            store.delete(obj)
    
    def batch_delete(self, objs):
        
        # delete-through
        for store in self.__stores:
            store.batch_delete(objs)
    
    def get(self, klass, key):
        
        # pull-through
        for i in range(0, len(self.__stores)):
            obj = self.__stores[i].get(klass, key)
            if obj:
                # we found it, pull it through all previous layers
                for previous_store in self.__stores[0:i]:
                    previous_store.put(obj)
                return obj
        
        # never found it
        return None
    
    def batch_get(self, klass, keys):
        
        # pull-through
        final_objects = [None for k in keys]
        for i in range(0, len(self.__stores)):
        
            # get as many hits as possible from this layer
            null_indices = [idx for idx in range(0, len(final_objects)) if final_objects[idx] is None]
            null_keys = [keys[null_idx] for null_idx in null_indices]
            new_objects = self.__stores[i].batch_get(klass, null_keys)
            nonnull_new_objects = []
            for null_idx, new_object in zip(null_indices, new_objects):
                final_objects[null_idx] = new_object
                if new_object:
                    nonnull_new_objects.append(new_object)
            
            # make sure any new objects are persisted in upper layers
            for previous_store in self.__stores[0:i]:
                previous_store.batch_put(nonnull_new_objects)
        
        # return whichever objects we found
        return final_objects
        


class KeyValueHelper(object):
    """Internal helper object that can generate unique keys for a store that
    stores objects in key/value pairs.  Given a class/instance and a property
    dictionary, this helper creates a unique lookup key (e.g. 'mymodule.MyClass(foo=abc;bar=123)')"""
    
    def __init__(self, polymorphic=False, prefix="stockpyle"):
        self.__stockpyle_bases_lookup = {}
        self.__polymorphic = polymorphic
        self.__prefix = prefix
    
    def generate_verbose_lookup_key(self, target_klass, properties):
        properties_str = self.__create_properties_str(properties)
        return "%s:%s.%s(%s)" % (self.__prefix, str(target_klass.__module__), target_klass.__name__, properties_str)

    def generate_terse_lookup_key(self, target_klass, properties):
        properties_str = self.__create_properties_str(properties)
        return "%s:%s.%s(%s)" % (self.__prefix, str(target_klass.__module__), target_klass.__name__, hash(properties_str))
    
    def generate_all_verbose_lookup_keys(self, obj):
        return self.__generate_all_lookup_keys(obj, self.generate_verbose_lookup_key)
        
    def generate_all_terse_lookup_keys(self, obj):
        return self.__generate_all_lookup_keys(obj, self.generate_terse_lookup_key)
    
    def __create_properties_str(self, properties):
        # FIXME: probably should use repr() for the values
        return ";".join(sorted(["%s=%s" % kv for kv in properties.iteritems()]))
    
    def __generate_all_lookup_keys(self, obj, generate_lookup_key_callback):
        lookup_keys = []
        klasses = [obj.__class__]
        if self.__polymorphic:
            klasses += self.__get_stockpyle_base_classes(obj.__class__)
        for klass in klasses:
            for stockpyle_key in klass.__stockpyle_keys__:
                if isinstance(stockpyle_key, tuple):
                    property_names = stockpyle_key
                else:
                    property_names = tuple([stockpyle_key])
                property_dict = self.__get_property_dict(obj, property_names)
                lookup_keys.append(generate_lookup_key_callback(klass, property_dict))
        return lookup_keys
    
    def __get_stockpyle_base_classes(self, klass):
        """returns an ordered list of stockpyle-managed base classes by recursing
        up the inheritance tree of the given class and collecting any base classes
        that have __stockpyle_keys__ defined"""
        if klass not in self.__stockpyle_bases_lookup:
            
            # we haven't calculated the stockpyle bases for this class yet
            # calculate them
            bases = []
            def collect(current_klass):
                for b in current_klass.__bases__:
                    if hasattr(b, "__stockpyle_keys__"):
                        bases.append(b)
                    collect(b)
            collect(klass)
            
            # and then save for for faster lookup later
            self.__stockpyle_bases_lookup[klass] = bases
            
        # return those bases
        return self.__stockpyle_bases_lookup[klass]
    
    def __get_property_dict(self, obj, property_names):
        """for a given object, returns a dictionary of all attributes
        with the given names"""
        properties = {}
        for n in property_names:
            properties[n] = getattr(obj, n)
        return properties



from unittest import TestCase
from datetime import datetime, timedelta
from tests.fixture_store import StoreFixture, ExpiringStoreFixture, Foobar
from tests.mocks import MockMemcacheClient
from stockpyle._memcache import MemcacheStore


class MemcacheStoreTestCase(StoreFixture, ExpiringStoreFixture, TestCase):
    
    store = property(lambda self: self.__mc)
    
    def setUp(self):
        self.__mc = MemcacheStore(client=MockMemcacheClient(), lifetime_cb=self.lifetime_cb)
    
    def test_must_specify_client_or_servers(self):
        self.assertRaises(ValueError, lambda: MemcacheStore())
        self.assertRaises(ValueError, lambda: MemcacheStore(client=MockMemcacheClient(), servers=["localhost:8080"]))
        try:
            MemcacheStore(servers=["localhost:8080"])
        except Exception, e:
            self.fail("should be able to create client against server list (%s)" % str(e))
    
    def test_integer_lifetime_must_be_within_memcache_api_limitation(self):
        max_int_delta = 60*60*24*30
        def lifetime_cb(obj):
            # return the 'foo' attribute as a lifetime
            return obj.foo
        mc = MemcacheStore(client=MockMemcacheClient(), lifetime_cb=lifetime_cb)
        obj1 = Foobar(max_int_delta, 2, 3)
        obj2 = Foobar(max_int_delta+1, 2, 3)
        obj3 = Foobar(max_int_delta-1, 2, 3)
        self.assertRaises(ValueError, lambda: mc.put(obj1))
        self.assertRaises(ValueError, lambda: mc.put(obj2))
        try:
            mc.put(obj3)
        except ValueError, e:
            self.fail("obj3 has a lifetime just within the API limit, this should have succeeded")
        
    def test_lifetime_must_be_datetime_timedelta_or_integer(self):
        def lifetime_cb(obj):
            # return the 'foo' attribute as a lifetime
            return obj.foo
        mc = MemcacheStore(client=MockMemcacheClient(), lifetime_cb=lifetime_cb)
        obj1 = Foobar(1.2, 2, 3)
        obj2 = Foobar('blah', 2, 3)
        obj3 = Foobar(555, 2, 3)
        obj4 = Foobar(timedelta(seconds=555), 2, 3)
        obj5 = Foobar(datetime.today(), 2, 3)
        self.assertRaises(ValueError, lambda: mc.put(obj1))
        self.assertRaises(ValueError, lambda: mc.put(obj2))
        try:
            mc.put(obj3)
        except ValueError, e:
            self.fail("obj3 has a lifetime of an integer, it should be just fine to put")
        try:
            mc.put(obj4)
        except ValueError, e:
            self.fail("obj4 has a lifetime of a timedelta, it should be just fine to put")
        try:
            mc.put(obj5)
        except ValueError, e:
            self.fail("obj4 has a lifetime of a datetime, it should be just fine to put")


import sys
import unittest
import logging


class _UnittestLogHandler(object):
    """acts as a StreamHandler on stdout (does some additional formatting)"""
    
    def __init__(self, log_name):
        self.__log_name = log_name

    def write(self, msg):
        sys.stdout.write(">>> %s: %s" % (self.__log_name, msg))
    
    def __getattr__(self, name):
        return getattr(sys.stdout, name)


def monitor_log(log_name):
    """monitors a log for unit tests"""
    logger = logging.getLogger(log_name)
    logger.addHandler(logging.StreamHandler(_UnittestLogHandler(log_name)))
    logger.setLevel(logging.INFO)


###############################################################################
#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

COVERAGE = [
    '--with-coverage',
    '--cover-package=stockpyle',
    ]

FLAGS = ["-v"] + COVERAGE# + ["--tests=test_sqlalchemy_store"]

#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
###############################################################################

if __name__ == "__main__":

    import sys
    import os
    import nose
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

    # run nosetests
    nose.core.TestProgram(defaultTest="tests", argv=sys.argv+FLAGS)


from datetime import datetime, timedelta

class MockMemcacheClient(object):
    """stubs out memcached calls with an in-memory compatible API"""
    
    def __init__(self):
        self.__map = {}
        self.__timeouts = {}
        self.flush_all()
    
    def get(self, key):
        now = datetime.today()
        if key in self.__timeouts:
            if now >= self.__timeouts[key]:
                del self.__timeouts[key]
                del self.__map[key]
                return None
        
        val = self.__map.get(key, None)
        if val:
            return val
        else:
            return None
    
    def set(self, key, value, time=None):
        
        # convert integer time into python time objects
        if time:
            if time < 60*60*24*30:
                delta = timedelta(seconds=time)
                time = datetime.today() + delta
            else:
                time = datetime.fromtimestamp(time)
            
        self.__map[key] = value
        if time:
            self.__timeouts[key] = time
        return True
    
    def get_multi(self, keys):
        object_map = {}
        for k in keys:
            object_map[k] = self.get(k)
        return object_map
    
    def set_multi(self, obj_map, time=None):
        for key, obj in obj_map.iteritems():
            self.set(key, obj, time)
    
    def delete_multi(self, keys):
        for k in keys:
            self.delete(k)
        
    def delete(self, key):
        if key in self.__map:
            del self.__map[key]
            if key in self.__timeouts:
                del self.__timeouts[key]
            return True
        else:
            return False
    
    def flush_all(self):
        self.__map = {}
        self.__timeouts = {}



try:
    import shove
    __HAVE_SHOVE = True
except ImportError:
    import sys
    print >> sys.stderr, "\n*** warning: skipping optional Shove tests, you don't have the module (http://pypi.python.org/pypi/shove)"
    __HAVE_SHOVE = False


if __HAVE_SHOVE:

    from unittest import TestCase
    from tests.fixture_store import StoreFixture, ExpiringStoreFixture
    from stockpyle._shove import ShoveStore


    class ShoveStoreTestCase(StoreFixture, ExpiringStoreFixture, TestCase):
    
        store = property(lambda self: self.__store)
    
        def setUp(self):
            from shove import Shove
            self.__inmemory_shove = Shove("simple://")
            self.__store = ShoveStore(shove=self.__inmemory_shove, lifetime_cb=self.lifetime_cb)
    
    
    class ShoveStoreViaUriTestCase(StoreFixture, ExpiringStoreFixture, TestCase):
    
        store = property(lambda self: self.__store)
    
        def setUp(self):
            self.__store = ShoveStore(shoveuri="simple://", lifetime_cb=self.lifetime_cb)
    
    
    class ShoveStoreIntantiationTestCase(TestCase):
        
        def test_requires_shove_or_shove_uri(self):
            
            self.assertRaises(ValueError, lambda: ShoveStore())
        
        def test_cannot_provide_shove_and_shove_uri_at_same_time(self):
            
            from shove import Shove
            s = Shove()
            self.assertRaises(ValueError, lambda: ShoveStore(shove=s, shoveuri="simple://"))
            


from datetime import datetime, timedelta
from stockpyle._helpers import KeyValueHelper

    
class BaseStore(object):
    """represents a standard storage container"""
        
    def put(self, obj):
        raise NotImplementedError()
        
    def batch_put(self, objs):
        raise NotImplementedError()
    
    def delete(self, obj):
        raise NotImplementedError()
    
    def batch_delete(self, objs):
        raise NotImplementedError()
    
    def get(self, klass, key):
        raise NotImplementedError()
    
    def batch_get(self, klass, keys):
        raise NotImplementedError()


class BaseDictionaryStore(BaseStore):
    """Base class for all dictionary-like stores.  Optional capabilities include
    polymorphic stores and object expiration times."""
        
    def __init__(self, dictionary, polymorphic=False, lifetime_cb=None):
        """create a new BaseDictionaryStore that stores data in the given
        dictionary or dictionary-like object.  Options:
        
        polymorphic: set this to true to replicate stored objects
                     under their base classes' __stockpyle_keys__
                     so that they can be queried via the base class
                     as well
        
        lifetime_cb: this callback takes an object and returns an absolute
                     expiration time as a datetime, or a relative expiration
                     as a timedelta or integer number of seconds
        """
        super(BaseDictionaryStore, self).__init__()
        self.__keyvalue_helper = KeyValueHelper(polymorphic=polymorphic)
        self.__dictionary = dictionary
        if lifetime_cb:
            self.__lifetime_cb = lifetime_cb
        else:
            self.__lifetime_cb = lambda obj: None
        
    def put(self, obj):
        """stores the given object, indexed by its __stockpyle_keys__ and with
        and expiration date determined by the lifetime_cb in the constructor"""
        # get the lifetime as a datetime
        lifetime = self.__lifetime_cb(obj)
        if lifetime and not isinstance(lifetime, datetime):
            if isinstance(lifetime, timedelta):
                lifetime = datetime.today() + lifetime
            elif isinstance(lifetime, int) or isinstance(lifetime, float):
                lifetime = datetime.today() + timedelta(seconds=int(lifetime))
        
        for k in self.__keyvalue_helper.generate_all_verbose_lookup_keys(obj):
            self.__dictionary[k] = (obj, lifetime)

    def batch_put(self, objs):
        """puts a list of objects into the dictionary store"""
        for obj in objs:
            self.put(obj)
    
    def delete(self, obj):
        """deletes an object from the dictionary store"""
        cachekeys = self.__keyvalue_helper.generate_all_verbose_lookup_keys(obj)
        for k in cachekeys:
            if k in self.__dictionary:
                del self.__dictionary[k]
    
    def batch_delete(self, objs):
        """deletes a list of objects from the dictionary store"""
        for o in objs:
            self.delete(o)
    
    def get(self, klass, key):
        """retrieves an object of the given class, under the given 'key' (a dictionary of key/value pairs)
        Expiration is checked before returning the object.  If the object has expired,
        this method will delete the object from the dictionary store."""
        cachekey = self.__keyvalue_helper.generate_verbose_lookup_key(klass, key)
        result = self.__dictionary.get(cachekey, None)
        if result:
            (obj, lifetime) = result
            if lifetime:
                if datetime.today() <= lifetime:
                    return obj
                else:
                    # expired
                    self.delete(obj)
                    return None
            else:
                # no lifetime, we don't have to worry about expiration
                return obj
        else:
            return None
    
    def batch_get(self, klass, keys):
        """returns a list of objects of the given class, under the given 'keys'
        (dictionaries of key/value pairs)"""
        return [self.get(klass, k) for k in keys]
        


from unittest import TestCase
from time import sleep
from collections import defaultdict
from tests.helpers import CountingStoreWrapper
from tests.fixture_store import StoreFixture, ExpiringStoreFixture
from tests.mocks import MockMemcacheClient
from stockpyle._chained import ChainedStore
from stockpyle._procmem import ProcessMemoryStore
from stockpyle._memcache import MemcacheStore


class Foobar(object):
    
    __stockpyle_keys__ = [("foo", "bar"), ("bar", "zap"), "zap"]
    
    def __init__(self, foo, bar, zap):
        self.foo = foo
        self.bar = bar
        self.zap = zap
    
    def __eq__(self, other):
        return other and (other.foo, other.bar == self.foo, self.bar)


class ChainedStoreWithLocalAndMemcacheTestCase(StoreFixture, ExpiringStoreFixture, TestCase):
    
    store = property(lambda self: self.__store)
    
    def setUp(self):
        local = ProcessMemoryStore(lifetime_cb=self.lifetime_cb)
        mc = MemcacheStore(client=MockMemcacheClient(), lifetime_cb=self.lifetime_cb)
        self.__store = ChainedStore([local, mc])
    
    def test_cannot_instantiate_with_no_stores(self):
        self.assertRaises(ValueError, lambda: ChainedStore())
        self.assertRaises(ValueError, lambda: ChainedStore([]))


class ChainedStoreWriteThroughTestCase(TestCase):
    
    def local_lifetime_cb(self, obj):
        if hasattr(obj.__class__, "__stockpyle_unittest_local_lifetime__"):
            return obj.__class__.__stockpyle_unittest_local_lifetime__
        else:
            return None
    
    def mc_lifetime_cb(self, obj):
        if hasattr(obj.__class__, "__stockpyle_unittest_mc_lifetime__"):
            return obj.__class__.__stockpyle_unittest_mc_lifetime__
        else:
            return None
    
    store = property(lambda self: self.__store)
    
    def setUp(self):
        self.__local = CountingStoreWrapper(ProcessMemoryStore(lifetime_cb=self.local_lifetime_cb))
        self.__mc = CountingStoreWrapper(MemcacheStore(client=MockMemcacheClient(), lifetime_cb=self.mc_lifetime_cb))
        self.__store = ChainedStore([self.__local, self.__mc])
    
    def test_after_writethrough_gets_from_only_local(self):
        obj = Foobar(1,2,3)
        
        self.__store.put(obj)
        
        obj_r = self.__store.get(Foobar, {"foo": 1, "bar": 2})
        self.assertEqual(obj_r, obj)
        
        self.assertEqual(self.__local.counts["get"], 1)
        self.assertEqual(self.__mc.counts["get"], 0)
        
    def test_can_expire_earlier_in_certain_caches(self):
        
        # lasts 1 second longer in memcached
        Foobar.__stockpyle_unittest_local_lifetime__ = 1
        Foobar.__stockpyle_unittest_mc_lifetime__ = 2
        
        try:
        
            obj = Foobar(1,2,3)
        
            self.__store.put(obj)
        
            obj_r = self.__store.get(Foobar, {"foo": 1, "bar": 2})
            self.assertEqual(obj_r, obj)
        
            self.assertEqual(self.__local.counts["get"], 1)
            self.assertEqual(self.__mc.counts["get"], 0)
        
            # now wait for the local cache to expire
            sleep(1)
        
            obj_r = self.__store.get(Foobar, {"foo": 1, "bar": 2})
            self.assertEqual(obj_r, obj)
        
            self.assertEqual(self.__local.counts["get"], 2)
            self.assertEqual(self.__mc.counts["get"], 1)
        
            # finally wait for all caches to expire
            sleep(2)
        
            obj_r = self.__store.get(Foobar, {"foo": 1, "bar": 2})
            self.assertEqual(obj_r, None, "object should be completely gone now")
        
            self.assertEqual(self.__local.counts["get"], 3)
            self.assertEqual(self.__mc.counts["get"], 2)
        
        finally:
            Foobar.__stockpyle_unittest_local_lifetime__ = None
            Foobar.__stockpyle_unittest_mc_lifetime__ = None




import time
from datetime import datetime, timedelta
from stockpyle._base import BaseStore
from stockpyle._helpers import KeyValueHelper


class MemcacheStore(BaseStore):
    
    def __init__(self, client=None, servers=None, polymorphic=False, lifetime_cb=None):
        super(MemcacheStore, self).__init__()
        
        # set up the lifetime callback and key/value generator
        self.__keyvalue_helper = KeyValueHelper(polymorphic=polymorphic)
        if lifetime_cb:
            self.__lifetime_cb = lifetime_cb
        else:
            self.__lifetime_cb = lambda obj: None
        
        # get/create the memcache client object
        if client and servers:
            raise ValueError("cannot give both a client and a list of servers, only one is allowed")
            
        if client:
            self.__client = client
        elif servers:
            import memcache
            self.__client = memcache.Client(servers)
        else:
            raise ValueError("you must specify either a client or a list of servers for memcache")
    
    def __get_memcache_compatible_lifetime(self, obj):
        """returns the lifetime for an object as an integer time since
        the epoch, or as a delta number of seconds (according to the memcache API
        limitation that deltas must be < 60*60*24*30 seconds)"""
        
        # get default lifetime
        lifetime = self.__lifetime_cb(obj)
        if lifetime is None:
            return 0
            
        else:
            # TODO: memoize this block of code for speed
            # ensure that lifetime is in unixtime or seconds
            if isinstance(lifetime, datetime):
            
                # unixtime
                return time.mktime(lifetime.timetuple())
            
            else:
                # timedelta
                if isinstance(lifetime, timedelta):
                    # convert to integer seconds
                    lifetime = int(86400*lifetime.days + lifetime.seconds + lifetime.microseconds/1000000.0)

                if isinstance(lifetime, int):
                    # check delta expiration against memcache API limitation
                    if lifetime >= 60*60*24*30:
                        raise ValueError("delta expiration must be less than 60*60*24*30 seconds (memcache API limitation)")
                else:
                    raise ValueError("lifetime must be an integer, datetime, or timedelta (got %s)" % lifetime.__class__)
                
                # everything checked out ok
                return lifetime
    
    def put(self, obj):
        lifetime = self.__get_memcache_compatible_lifetime(obj)
        object_map = {}
        for k in self.__keyvalue_helper.generate_all_terse_lookup_keys(obj):
            object_map[k] = obj
        self.__client.set_multi(object_map, lifetime)
        
    def batch_put(self, objs):
        
        # group puts by lifetime
        lifetime_lookup = {}
        for obj in objs:
            lifetime = self.__get_memcache_compatible_lifetime(obj)
            if lifetime not in lifetime_lookup:
                lifetime_lookup[lifetime] = []
            lifetime_lookup[lifetime].append(obj)
        
        # for each lifetime, do a set_multi to drop those objects into memcache
        # with that same expiration time
        for lifetime in lifetime_lookup:
            objs_with_same_lifetime = lifetime_lookup[lifetime]
            object_map = {}
            for obj in objs_with_same_lifetime:
                for k in self.__keyvalue_helper.generate_all_terse_lookup_keys(obj):
                    object_map[k] = obj
            self.__client.set_multi(object_map, lifetime)
    
    def delete(self, obj):
        cachekeys = self.__keyvalue_helper.generate_all_terse_lookup_keys(obj)
        self.__client.delete_multi(cachekeys)
    
    def batch_delete(self, objs):
        cachekeys = []
        for obj in objs:
            cachekeys += self.__keyvalue_helper.generate_all_terse_lookup_keys(obj)
        self.__client.delete_multi(cachekeys)
    
    def get(self, klass, key):
        cachekey = self.__keyvalue_helper.generate_terse_lookup_key(klass, key)
        return self.__client.get(cachekey)
    
    def batch_get(self, klass, keys):
        cachekeys = [self.__keyvalue_helper.generate_terse_lookup_key(klass, key) for key in keys]
        object_map = self.__client.get_multi(cachekeys)
        return [object_map.get(k, None) for k in cachekeys]

            


import threading
from stockpyle._base import BaseStore
try:
    from sqlalchemy.orm import EXT_CONTINUE, SessionExtension
    __HAS_SQLALCHEMY = True
except ImportError:
    __HAS_SQLALCHEMY = False
    

if __HAS_SQLALCHEMY:


    # SQLAlchemy is installed
    class StockpyleSessionExtension(SessionExtension):
        """If you have your own sessions, you can add this extension
        to force that session to synchronize objects with the given store"""
        
        def __init__(self, store):
            self.__store = store
            self.__operation_for_instance = {}
        
        def after_flush(self, session, flush_context):
            
            # deleted items must be deleted from the store
            for o in session.deleted:
                self.__operation_for_instance[o] = "delete"
            
            # new items must be put to the store
            for o in session.new:
                self.__operation_for_instance[o] = "put"
            
            # dirty items must be put to the store as well
            for o in session.dirty:
                self.__operation_for_instance[o] = "put"
                
            return EXT_CONTINUE
        
        def after_commit(self, session):
            
            # group operations by deletes/puts
            operation_map = {}
            for obj, operation in self.__operation_for_instance.iteritems():
                if operation not in operation_map:
                    operation_map[operation] = []
                operation_map[operation].append(obj)
            
            # do a batch delete for all deletes
            if "delete" in operation_map:
                self.__store.batch_delete(operation_map["delete"])
            
            # do a batch put for all puts
            if "put" in operation_map:
                self.__store.batch_put(operation_map["put"])
            
            # clear the operation cache, we're done
            self.__operation_for_instance = {}
            
            return EXT_CONTINUE
            
            
else:
    
    
    # SQLAlchemy is not installed
    class StockpyleSessionExtension(object):
        
        def __init__(self):
            raise StandardError("cannot use StockpyleSessionExtension without the SQLAlchemy module")


def _is_sqlalchemy_class(klass):
    """returns true if the given class is managed by SQLAlchemy"""
    # TODO: better way?
    return hasattr(klass, "_sa_class_manager")

def _is_sqlalchemy_object(obj):
    """returns true if the given object is managed by SQLAlchemy"""
    # TODO: better way?
    return hasattr(obj, "_sa_instance_state")


class SqlAlchemyStore(BaseStore):
    """Represents a storage target that can store any SQLAlchemy-mapped object.
    Non-SQLAlchemy objects will be ignored by this store."""
    
    def __init__(self, uri=None, engine=None, **kwargs):
        """Create a new SqlAlchemyStore.  When given a connection string (uri)
        or SQLAlchemy Engine (engine), this Store will create it's own internal
        SQLAlchemy Session to manage objects.  If you do not provide a URI or
        Engine, your mapped object metadata must be bound to their own engines.
        """

        super(SqlAlchemyStore, self).__init__()
        
        # get the session
        if "session" in kwargs:
            # we no longer allow initialization with a pre-existing session object, there
            # are too many issues with this approach at the moment
            raise DeprecationWarning("cannot instantiate a SqlAlchemyStore with a pre-existing session object")
        else:
            # no session, we have to make one
            # first we need to get the engine
            from sqlalchemy.orm import sessionmaker
            
            if uri and engine:
                # can't give two ways to get an engine
                raise ValueError("you can only provide either a connection string URI or an engine, not both")
                
            elif uri:
                # we have a uri to create an engine
                from sqlalchemy import create_engine
                engine = create_engine(uri)
            
            if engine:
                # we have an engine, we can create the bound session now
                self.__session = sessionmaker(autoflush=True, bind=engine)()
            
            else:
                # no engine or URI was specified, create an unbound session
                # (mapped object metadata will need to be bound to an engine in this case)
                self.__session = sessionmaker(autoflush=True)()
    
    def put(self, obj):
        # only store SA objects
        if _is_sqlalchemy_object(obj):
            self._sa_put(obj)
        
    def batch_put(self, objs):
        # only store SA objects
        sa_objs = [o for o in objs if _is_sqlalchemy_object(o)]
        if sa_objs:
            self._sa_batch_put(sa_objs)
    
    def delete(self, obj):
        # only delete SA objects
        if _is_sqlalchemy_object(obj):
            self._sa_delete(obj)
    
    def batch_delete(self, objs):
        # only delete SA objects
        sa_objs = [o for o in objs if _is_sqlalchemy_object(o)]
        if sa_objs:
            self._sa_batch_delete(sa_objs)
    
    def get(self, klass, key):
        # only attempt to query SQAlchemy objects
        if _is_sqlalchemy_class(klass):
            return self._sa_get(klass, key)
        else:
            return None
    
    def batch_get(self, klass, keys):
        # only attempt to query SQAlchemy objects
        if _is_sqlalchemy_class(klass):
            return self._sa_batch_get(klass, keys)
        else:
            return [None for k in keys]
    
    def _sa_put(self, obj):
        from sqlalchemy.exceptions import InvalidRequestError
        try:
            self.__session.add(obj)
            self.__session.commit()
        except InvalidRequestError:
            # this is likely due to mismatched session attachment, do the more expensive 'merge' operation
            try:
                self.__session.merge(obj)
                self.__session.commit()
            except:
                self.__session.rollback()
                raise

    def _sa_batch_put(self, objs):
        from sqlalchemy.exceptions import InvalidRequestError
    
        # commit the session
        try:
            self.__session.add_all(objs)
            self.__session.commit()
        except InvalidRequestError:
            # this is likely due to mismatched session attachment, do the more expensive 'merge' operation
            try:
                for obj in objs:
                    self.__session.merge(obj)
                self.__session.commit()
            except:
                self.__session.rollback()
                raise
    
    def _sa_delete(self, obj):
        from sqlalchemy.exceptions import InvalidRequestError
        try:
            self.__session.delete(obj)
            self.__session.commit()
        except InvalidRequestError:
            # this is likely due to mismatched session attachment, do the more expensive 'merge' operation
            try:
                obj = self.__session.merge(obj)
                self.__session.delete(obj)
                self.__session.commit()
            except:
                self.__session.rollback()
                raise
    
    def _sa_batch_delete(self, objs):
        from sqlalchemy.exceptions import InvalidRequestError

        # TODO: is there a bulk_delete in SA?
        try:
            for obj in objs:
                self.__session.delete(obj)
            self.__session.commit()
        except InvalidRequestError:
            # this is likely due to mismatched session attachment, do the more expensive 'merge' operation
            try:
                for obj in objs:
                    obj = self.__session.merge(obj)
                    self.__session.delete(obj)
                self.__session.commit()
            except:
                self.__session.rollback()
                raise
    
    def _sa_get(self, klass, key):
        return self.__session.query(klass).filter_by(**key).first()
    
    def _sa_batch_get(self, klass, keys):
        # FIXME: this is very inefficient, need to read up on the SA API a bit more
        return [self.get(klass, k) for k in keys]
        


try:
    import sqlalchemy
    __HAVE_SA = True
except ImportError:
    import sys
    print >> sys.stderr, "\n*** warning: skipping optional SQLAlchemy tests, you don't have the module (http://pypi.python.org/pypi/SQLAlchemy)"
    __HAVE_SA = False


if __HAVE_SA:

    from unittest import TestCase
    from sqlalchemy import Column, create_engine
    from sqlalchemy.types import Integer, String
    from sqlalchemy.orm import sessionmaker
    from tests.helpers import CountingStoreWrapper
    from tests.fixture_store import StoreFixture, DeclarativeBase, Foobar
    from stockpyle._chained import ChainedStore
    from stockpyle._procmem import ProcessMemoryStore
    from stockpyle._sqlalchemy import SqlAlchemyStore, StockpyleSessionExtension
    
    
    class Bazbot(DeclarativeBase):

        __tablename__ = "bazbots"
    
        zap = Column(Integer, primary_key=True)
        foo = Column(Integer)
        bar = Column(Integer)

        __stockpyle_keys__ = [("foo", "bar"), ("bar", "zap"), "zap"]

        def __init__(self, foo, bar, zap):
            self.foo = foo
            self.bar = bar
            self.zap = zap

        def __eq__(self, other):
            return other and (other.foo, other.bar == self.foo, self.bar)


    class SqlAlchemyStoreTestCase(StoreFixture, TestCase):
    
        store = property(lambda self: self.__store)
    
        def setUp(self):
            self.__engine = create_engine("sqlite:///:memory:")
            self.__store = SqlAlchemyStore(engine=self.__engine)
            DeclarativeBase.metadata.create_all(bind=self.__engine)
        
        def tearDown(self):
            DeclarativeBase.metadata.drop_all(bind=self.__engine)
        
        def test_instantiation_with_session_generates_deprecation_warning(self):
            session = sessionmaker(autoflush=True, bind=self.__engine)()
            self.assertRaises(DeprecationWarning, lambda: SqlAlchemyStore(session=session))
        
        def test_instantiation_with_uri_works(self):
            store = SqlAlchemyStore(uri="sqlite:///:memory:")
        
        def test_instantiation_with_nothing_is_okay_and_requires_objects_to_be_bound_to_engine(self):
            # create an engine-unbound SqlAlchemyStore
            store = SqlAlchemyStore()
            obj = Foobar(1,2,3)
            self.assertRaises(Exception, lambda: store.put(obj))
            DeclarativeBase.metadata.bind = self.__engine
            try:
                # now we should be able to put the object
                store.put(obj)
            finally:
                DeclarativeBase.metadata.bind = None
        
        def test_instantiation_with_connstring_and_engine_raises_error(self):
            engine = create_engine("sqlite:///:memory:")
            self.assertRaises(ValueError, lambda: SqlAlchemyStore(uri="sqlite:///:memory:", engine=engine))
        
        def test_ignores_non_sqlalchemy_mapped_objects(self):
            class NonMappedObj(object):
                __stockpyle_keys__ = ["mykey"]
                mykey = 123
            
            obj = NonMappedObj()
            self.__store.put(obj)
            
            self.__store.batch_put([obj])

            obj_r = self.__store.get(NonMappedObj, {"mykey": 123})
            self.assertEqual(obj_r, None)

            obj_list = self.__store.batch_get(NonMappedObj, [{"mykey": 123}])
            self.assertEqual(obj_list, [None])
            
            self.__store.batch_delete([obj])
        
        def test_external_session_must_merge_objects_retrieved_from_stockpyle(self):

            obj = Foobar(1,2,3)
            self.store.put(obj)
            
            # create an external session
            session = sessionmaker(autoflush=True, bind=self.__engine)()
            
            # change the object, we're going to save it to SA directly
            obj.foo = 999
        
            # try to add to SA session, should fail
            self.assertRaises(Exception, lambda: session.add(obj))
            
            # merge instead
            session.merge(obj)
            session.commit()
            
            # now that should be in SA with the new 999 value for obj.foo
            obj_r = session.query(Foobar).first()
            self.assertEqual(obj_r, obj)
            self.assertEqual(obj_r.foo, 999)

        def test_can_delete_retrieved_object_from_external_session(self):
        
            obj = Foobar(1,2,3)
            self.store.put(obj)
        
            # get using a sqlalchemy session
            session = sessionmaker(autoflush=True, bind=self.__engine)()
            obj_r = session.query(Foobar).first()
            self.assertEqual(obj_r, obj)
        
            # now delete the obj through the store
            self.store.delete(obj_r)
        
            # store should say empty
            obj_r = self.store.get(Foobar, {"foo": 1, "bar": 2})
            self.assertEqual(obj_r, None)
        
            # raw db request should say empty
            obj_r = session.query(Foobar).first()
            self.assertEqual(obj_r, None)
        
        def test_can_do_batch_delete_on_retrieved_objects_from_external_session(self):
        
            obj1 = Foobar(1,2,3)
            obj2 = Foobar(4,5,6)
        
            self.store.put(obj1)
            self.store.put(obj2)
        
            # get using a sqlalchemy session
            session = sessionmaker(autoflush=True, bind=self.__engine)()
            objs = session.query(Foobar).all()
            self.assertEqual(len(objs), 2)
        
            # verify we got the same objects out
            obj1_r, obj2_r = tuple(objs)
            self.assertEqual(obj1_r, obj1)
            self.assertEqual(obj2_r, obj2)
        
            # now delete them
            # session.expunge(obj1_r)
            # session.expunge(obj2_r)
            self.store.batch_delete([obj1_r, obj2_r])
        
            obj1_r, obj2_r = self.store.batch_get(Foobar, [
                {"foo": 1, "bar": 2}, #obj1
                {"foo": 4, "bar": 5}, #obj2
                ])
        
            self.assertEqual(obj1_r, None)
            self.assertEqual(obj2_r, None)

        def test_can_put_retrieved_object_from_external_session(self):
        
            obj = Foobar(1,2,3)
            self.store.put(obj)
        
            # get using a sqlalchemy session
            session = sessionmaker(autoflush=True, bind=self.__engine)()
            obj_r = session.query(Foobar).first()
            self.assertEqual(obj_r, obj)
        
            # now put the obj through the store
            self.store.put(obj_r)
        
            # store should give obj back
            obj_r = self.store.get(Foobar, {"foo": 1, "bar": 2})
            self.assertEqual(obj_r, obj)
        
            # raw db request should say the same
            obj_r = session.query(Foobar).first()
            self.assertEqual(obj_r, obj)

        def test_can_batch_put_retrieved_object_from_external_session(self):
        
            obj = Foobar(1,2,3)
            self.store.put(obj)
        
            # get using a sqlalchemy session
            session = sessionmaker(autoflush=True, bind=self.__engine)()
            obj_r = session.query(Foobar).first()
            self.assertEqual(obj_r, obj)
        
            # now put the obj through the store
            self.store.batch_put([obj_r])
        
            # store should give objs back
            obj_r = self.store.get(Foobar, {"foo": 1, "bar": 2})
            self.assertEqual(obj_r, obj)
        
            # raw db request should say the same
            obj_r = session.query(Foobar).first()
            self.assertEqual(obj_r, obj)


    class ChainedStoreWithSqlAlchemyTestCase(TestCase):
    
        def setUp(self):
            self.__lm = CountingStoreWrapper(ProcessMemoryStore())
            self.__engine = create_engine("sqlite:///:memory:")
            self.__sa = CountingStoreWrapper(SqlAlchemyStore(engine=self.__engine))
            self.__store = ChainedStore([self.__lm, self.__sa])
            
            # create a special synchronizing session
            ext = StockpyleSessionExtension(self.__lm)
            self.__session = sessionmaker(autoflush=True, extension=ext, bind=self.__engine)()
            
            DeclarativeBase.metadata.create_all(bind=self.__engine)

        def tearDown(self):
            DeclarativeBase.metadata.drop_all(bind=self.__engine)
    
        def test_local_memory_updated_even_when_object_updated_outside_stockpyle(self):
        
            # create
            obj = Bazbot(1,2,3)
            
            # put in the store (sanity check)
            self.assertEqual(self.__sa.counts["put"], 0)
            self.__store.put(obj)
            self.assertEqual(self.__sa.counts["put"], 1)
            
            # retrieve from the store (sanity check)
            obj_r = self.__store.get(Bazbot, {"zap": 3})
            self.assertEqual(obj_r, obj)
        
            # ensure the get was done from localmem, not SA
            self.assertEqual(self.__lm.counts["get"], 1)
            self.assertEqual(self.__sa.counts["get"], 0)
        
            # now, update obj outside of stockpyle
            sa_obj = self.__session.query(Bazbot).filter_by(zap=3).first()
            sa_obj.bar = 555
            self.__session.add(sa_obj)
        
            # the commit should cause the cache to be updated
            self.assertEqual(self.__lm.counts["batch_put"], 0)
            print "about to commit..."
            self.__session.commit()
            print "committed..."
            self.assertEqual(self.__lm.counts["batch_put"], 1)
            self.assertEqual(self.__sa.counts["put"], 1)
            
            # we can retrieve this object now from the cache,
            # even though it was updated
            obj_r = self.__lm.get(Bazbot, {"zap": 3})
            self.assertEqual(obj_r, obj)
        
            # ensure we still haven't ever done a SA persistent get
            self.assertEqual(self.__sa.counts["get"], 0)

        def test_local_memory_deleted_even_when_object_deleted_outside_stockpyle(self):
        
            # create
            obj = Bazbot(1,2,3)
            
            # put in the store (sanity check)
            self.assertEqual(self.__sa.counts["put"], 0)
            self.__store.put(obj)
            self.assertEqual(self.__sa.counts["put"], 1)
            
            # retrieve from the store (sanity check)
            obj_r = self.__store.get(Bazbot, {"zap": 3})
            self.assertEqual(obj_r, obj)

            # ensure the get was done from localmem, not SA
            self.assertEqual(self.__lm.counts["get"], 1)
            self.assertEqual(self.__sa.counts["get"], 0)
        
            # now, delete obj outside of stockpyle
            sa_obj = self.__session.query(Bazbot).filter_by(zap=3).first()
            self.__session.delete(sa_obj)
        
            # the commit should cause the cache to be updated
            self.assertEqual(self.__lm.counts["batch_delete"], 0)
            print "about to commit..."
            self.__session.commit()
            print "committed..."
            self.assertEqual(self.__lm.counts["batch_delete"], 1)
            self.assertEqual(self.__sa.counts["put"], 1)
            
            # cache will not have this object anymore
            obj_r = self.__lm.get(Bazbot, {"zap": 3})
            self.assertEqual(obj_r, None)
            
            # ensure we try the persistence layer and get what we want
            # when using the writethru
            obj_r = self.__store.get(Bazbot, {"zap": 3})
            self.assertEqual(self.__lm.counts["get"], 3)
            self.assertEqual(self.__sa.counts["get"], 1)
            self.assertEqual(obj_r, None)
            
        def test_local_memory_updated_even_when_object_created_outside_stockpyle(self):
        
            # create
            obj = Bazbot(1,2,3)
            
            # save obj outside of stockpyle
            self.__session.add(obj)
            
            # the commit should cause the cache to be updated
            self.assertEqual(self.__lm.counts["batch_put"], 0)
            print "about to commit..."
            self.__session.commit()
            print "committed..."
            self.assertEqual(self.__lm.counts["batch_put"], 1)
            self.assertEqual(self.__sa.counts["put"], 0)
            
            # we can retrieve this object now from the cache,
            # even though it was created outside of stockpyle
            obj_r = self.__lm.get(Bazbot, {"zap": 3})
            self.assertEqual(obj_r, obj)
        
            # ensure we still haven't ever done a SA persistent get
            self.assertEqual(self.__sa.counts["get"], 0)
            
            # but the persistent get should also work
            obj_r = self.__sa.get(Bazbot, {"zap": 3})
            self.assertEqual(obj_r, obj)



from collections import defaultdict


class CountingStoreWrapper(object):
    
    counts = property(lambda self: self.__counts)
    
    def __init__(self, store):
        self.__store = store
        self.__counts = defaultdict(lambda: 0)
    
    def __getattr__(self, name):
        self.__counts[name] += 1
        return getattr(self.__store, name)
